import os
import sys
import time
from pydub import AudioSegment as audio

# need to: create mfcc function, train/test set functions

# compute mfcc features using librosa
def mfcc(y, sr, num_mfcc, hop_length, delta):
    if delta = 0:
        return librosa.feature.mfcc(y=y, sr=sr, n_mfcc=num_mfcc, hop_length=hop_length)
    elif delta = 1:
        mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=num_mfcc, hop_length=hop_length)
        delta = librosa.feature.delta(mfcc)
        return np.c_[mfcc,delta]
    else:
        mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=num_mfcc, hop_length=hop_length)
        delta = librosa.feature.delta(mfcc)
        delta2 = librosa.feature.delta(mfcc, order=2)
        return np.c_[mfcc, delta, delta2]

# segment data function
def segment(data, seg_location, length):
  os.chdir(data)
  files = os.listdir(data)
  speakers = speech_data.get_speakers(data)
  waves = []
  num = {}
  for s in speakers:
    num[s] = 0
  c = 0
  for f in files: # grab all wave files in list
    waves.append(audio.from_wav(f))
    c = c + 1
  os.chdir(seg_location)
  for f,w in zip(files,waves): # need to segment the data into one second intervals
    begin = 0
    end = 1
    while (end*length) < int(w.duration_seconds):
      segment = w[begin*1000*length:end*1000*length]
      segment.export(speech_data.speaker(f) + '_' +  str(num[speech_data.speaker(f)]) + '.wav', 'wav')
      begin = begin + length
      end = end + length
      num[speech_data.speaker(f)] = num[speech_data.speaker(f)] + 1

# get speaker name from file
def speaker(file):
  return file.split("_")[0]
  #return file.split("_")[1]

# get speakers for labels in a directory
def get_speakers(path):
  files = os.listdir(path)
  print('number of files: %s' %len(files))
  def nobad(file):
    return "_" in file and not "." in file.split("_")[0]
  speakers=list(set(map(speaker,filter(nobad,files))))
  print(len(speakers)," speakers: ",speakers)
  return speakers

def one_hot_to_item(hot, items):
  i=np.argmax(hot)
  item=items[i]
  return item

def one_hot_from_item(item, items):
  x=[0]*len(items)
  i=items.index(item)
  x[i]=1
  return x



import os
import sys
import librosa
import pydub
import pyaudio
import tflearn
import audio_manip

'''
Need to figure out how to make this a function.
'''

# for now - directories
training_dir = sys.argv[1]
training_seg = sys.argv[2]

def create_net(num_layers, dropout, size, learning_rate, activation):
    # need to create input for shape of data
    # is this even necessary??

# train model over training directory, make sure to segment data
# if there is nothing in training_dir, need to record audio before segmenting?
if not os.listdir(training_dir):
    audio_manip.enroll(num_speakers=2, num_files=4, directory=training_dir)

length = 1
if not os.listdir(training_seg):
    audio_manip.segment(data=training_dir, seg_location=training_seg, length=length)

# create mfcc, delta, delta2 input features
hop_length = 128 # default = 512
X = []
Y = []
speakers = audio_manip.get_speakers(training_dir)
for f in os.listdir(training_seg):
    Y.append(audio_manip.one_hot_from_item(audio_manip.speaker(f), speakers))
    y, sr = librosa.load(training_seg + f)
    X.append(audio_manip.mfcc(y=y, sr=sr, num_mfcc=20, hop_length=128, delta=2)

# define the network for training
layer_size = 128
dropout = 0.7
learning_rate = 0.001
tflearn.init_graph(num_cores=8,gpu_memory_fraction=0.5)

net = tflearn.input_data(shape=[None, 20, 519])
net = tflearn.fully_connected(net, layer_size, activation='relu')
net = tflearn.fully_connected(net, layer_size, activation='relu')
net = tflearn.fully_connected(net, layer_size, activation='relu')
net = tflearn.dropout(net, dropout)
net = tflearn.fully_connected(net, len(speakers), activation='softmax')
net = tflearn.regression(net, optimizer='adam', loss='categorical_crossentropy', learning_rate=learning_rate)

# now train the model!
run_id='Demonstration'
model = tflearn.DNN(net)
model.fit(X, Y, n_epoch=50, show_metric=True, snapshot_step=1000, run_id=run_id, validation_set=0.1)

# save the model to model_dir
model.save(model_dir + 'demo.model')



import os
import sys
import librosa
import pydub
import pyaudio
import wave
import tflearn
import time
from sys import byteorder
from array import array
from struct import pack 

#constants
THRESHOLD = 500
CHUNK_SIZE = 1024
FORMAT = pyaudio.paInt16
RATE = 44100

def press_enter():
    while True:
        raw_input('Press enter to stop recording.')
        break
    return True

def is_silent(snd_data):
    # Returns 'True' if below the 'silent' threshold
    return max(snd_data) < THRESHOLD

def normalize(snd_data):
    # Average the volume out
    MAXIMUM = 16384
    times = float(MAXIMUM)/max(abs(i) for i in snd_data)

    r = array('h')
    for i in snd_data:
        r.append(int(i*times))
    return r

def trim(snd_data):
    # Trim the blank spots at the start and end
    def _trim(snd_data):
        snd_started = False
        r = array('h')

        for i in snd_data:
            if not snd_started and abs(i)>THRESHOLD:
                snd_started = True
                r.append(i)

            elif snd_started:
                r.append(i)
        return r

def add_silence(snd_data, seconds):
    # Add silence to the start and end of 'snd_data' of length 'seconds' (float)
    r = array('h', [0 for i in xrange(int(seconds*RATE))])
    r.extend(snd_data)
    r.extend([0 for i in xrange(int(seconds*RATE))])
    return r

def record():
    """
    Record a word or words from the microphone and 
    return the data as an array of signed shorts.
    Normalizes the audio, trims silence from the 
    start and end, and pads with 0.2 seconds of 
    blank sound.
    """
    p = pyaudio.PyAudio()
    stream = p.open(format=FORMAT, channels=1, rate=RATE,
        input=True, output=True,
        frames_per_buffer=CHUNK_SIZE)

    num_silent = 0
    snd_started = False

    r = array('h')

    while 1:
        # little endian, signed short
        snd_data = array('h', stream.read(CHUNK_SIZE))
        if byteorder == 'big':
            snd_data.byteswap()
        r.extend(snd_data)

        silent = is_silent(snd_data)

        if silent and snd_started:
            num_silent += 1
        elif not silent and not snd_started:
            snd_started = True
        # this controls the length of recording?
        if snd_started and press_enter():
            break

    sample_width = p.get_sample_size(FORMAT)
    stream.stop_stream()
    stream.close()
    p.terminate()

    r = normalize(r)
    r = trim(r)
    r = add_silence(r, 0.5) # does this have to be 0.5??
    return sample_width, r

def record_to_file(path):
    # Records from the microphone and outputs the resulting data to 'path'
    sample_width, data = record()
    data = pack('<' + ('h'*len(data)), *data)

    wf = wave.open(path, 'wb')
    wf.setnchannels(1)
    wf.setsampwidth(sample_width)
    wf.setframerate(RATE)
    wf.writeframes(data)
    wf.close()

# record audio for training, specify how many speakers and files per speaker with command line
# need to fix the time per recording - play with record_audio.record_to_file function
# maybe: Enter to begin, Enter to end?

def enroll(num_speakers, num_files, directory):
    count = 0
    while count < num_speakers:
        # record audio for each speaker - input strings for their names later
        for s in range(num_files):
            # record audio, save to a file
            raw_input('Press Enter to record an audio file.')
            t0 = time.time()
            record_to_file(directory + 'Speaker' + str(count) + '_' + str(s) + '.wav')
            print('Time during recording: %s' %str(time.time() - t0))
        count = count + 1

if __name__ == '__main__':
    num_speakers = int(sys.argv[1])
    num_files = int(sys.argv[2])
    directory = sys.argv[3]
    enroll(num_speakers=num_speakers, num_files=num_files, directory=directory)



import os
import sys
import librosa
import pydub
import pyaudio
import tflearn
import audio_manip

'''
Need to turn into a function that can be used.
'''

# directories - for now
testing_dir = sys.argv[1]
testing_seg = sys.argv[2]

# record for testing model - for now just doing one speaker test file, later create another python script for this
# record audio, save to a file
raw_input('Press Enter to record an audio file.')
record_audio.record_to_file(testing_dir + 'test.wav')

# segment the recorded test audio
if not os.listdir(testing_seg):
    segment_data.segment(data=testing_dir, seg_location=testing_seg, length=length)

# create mfcc, delta, delta2 for testing data
Xtest = []
for f in os.listdir(testing_seg):
    y, sr = librosa.load(testing_seg + f)
    Xtest.append(audio_manip.mfcc(y=y, sr=sr, num_mfcc=20, hop_length=128, delta=2)

# load the trained model from model_dir
model = DNN() # make sure this is correct
model.load(model_dir + 'demo.model') 

result = model.predict(Xtest)
c = 0
for f,r in zip(os.listdir(testing_seg), result):
    res = speech_data.one_hot_to_item(r, speakers)
    if res in f:
        c = c + 1
acc = float(c) / float(len(Xtest))
print('Test set accuracy: %s' %str(acc))

# need to add a part that outputs the speaker to command line



from setuptools import setup
from os import path


setup(

    name='pyspeakrec',

    version='0.1.0',

    description='A Speaker Recognition tool in Python',

    url='https://github.com/AKBoles/pyspeakrec',

    author='Andrew Boles',

    author_email='andrew.boles@my.utsa.edu',

    license='MIT',

    packages=['pyspeakrec'],

    install_requires=['tflearn','librosa','pydub','pyaudio'],

    keywords='speaker recognition, machine learning',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Programming Language :: Python :: 2.7'
    ],
    zip_safe=False
)
